===================================
Some Background about this solution
===================================

In Haskell object algebras can be implemented with type classes. In fact, 
precursors of the work on Object algebras can be found in a string of papers 
(see more details in the Object Algebras paper), starting from:

Generics for the Masses
Ralf Hinze
In Proceedings of the International Conference on Functional Programming (2004) 

Later papers ("Typecase: A design pattern for type-indexed functions" (Haskell 2005)
and the JFP version of "Generics for the Masses") explain the relation to encodings 
of datatypes. However, modularity and extensibility and the relation to the expression problem 
were only described later in the paper:

"Extensible and Modular Generics for the Masses"
Bruno C. d. S. Oliveira, Ralf Hinze and Andres Loeh
In Trends in Functional Programming (2006) 

The solution to the expression problem presented here is a (much simpler) version of
what's in that paper. Basically the details of the generic programming library complicate 
the code; arithmetic expressions are a lot simpler to encode. 

===================
The actual solution
===================

> {-# OPTIONS -XNoMonomorphismRestriction -XTypeSynonymInstances #-}

> module OA where

> import Prelude hiding (print)

Initial object algebra interface for expressions: integers and addition

> class ExpAlg e where
>   lit :: Int -> e
>   add :: e -> e -> e

An object algebra implementing that interface (evaluation)

> newtype Eval = Eval {eval :: Int} -- Evaluation interface
>
> instance ExpAlg Eval where  -- object algebra
>   lit        = Eval
>   add e1 e2  = Eval (eval e1 + eval e2) 

Evolution 1: Adding subtraction

> class ExpAlg e => SubExpAlg e where
>   sub :: e -> e -> e
>
> instance SubExpAlg Eval where
>   sub e1 e2 = Eval (eval e1 - eval e2)
	
Evolution 2: Adding pretty printing

> newtype Print = Print {print :: String}
>
> instance ExpAlg Print where
>   lit        = Print . show
>   add e1 e2  = Print (print e1 ++ " + " ++ print e2)
>
> instance SubExpAlg Print where
>   sub e1 e2 = Print (print e1 ++ " - " ++ print e2) 

Alternative Printer

> instance ExpAlg String where
>   lit        = show
>   add e1 e2  = e1 ++ " + " ++ e2
>
> instance SubExpAlg String where
>   sub e1 e2 = e1 ++ " - " ++ e2 

> exp1 = add (lit 3) (lit 4)
>
> exp2 = sub exp1 (lit 4)

In Haskell we exploit the type-directed instantiation mechanism of type classes, 
so we do not need to explicitly instanciate the object algebras (that is, 
the values "ea", "esa", "pa" and "pa2" in the Java solution).

> ev :: Eval 
> ev = exp1  -- automatic instantiation based on the type

Testing the actual algebras

> main = do  putStrLn ("Evaluation of exp1 \"" ++ print exp1 ++ "\" is: " ++ show (eval ev))
>            putStrLn ("Evaluation of exp1 \"" ++ print exp2 ++ "\" is: " ++ show (eval exp2))
>            putStrLn ("Evaluation of exp1 \"" ++ exp1 ++ "\" is: " ++ exp2)
